#version 450

#if defined(VULKAN) || defined(GL_core_profile) || defined(GL_ES)

	#define int4 ivec4
	#define uint2 uvec2
	#define float2 vec2
	#define float4 vec4

	#if defined(GL_ES)

		precision highp float;
		precision highp int;

	#endif

#elif defined(__METAL_VERSION__)

	#include <metal_geometric>

	using namespace metal;

#endif

#if !defined(__METAL_VERSION__)

	void SlugUnpack(float4 tex, float4 bnd, out float4 vbnd, out int4 vgly)

#else

	void SlugUnpack(float4 tex, float4 bnd, thread float4& vbnd, thread int4& vgly)

#endif

{
	#if defined(VULKAN) || defined(GL_core_profile) || defined(GL_ES)

		uint2 g = floatBitsToUint(tex.zw);

	#elif defined(__METAL_VERSION__)

		uint2 g = as_type<uint2>(tex.zw);

	#else

		uint2 g = asuint(tex.zw);

	#endif

	vgly = int4(g.x & 0xFFFFU, g.x >> 16U, g.y & 0xFFFFU, g.y >> 16U);
	vbnd = bnd;
}

#if !defined(__METAL_VERSION__)

	float2 SlugDilate(float4 pos, float4 tex, float4 jac, float4 m0, float4 m1, float4 m3, float2 dim, out float2 vpos)

#else

	float2 SlugDilate(float4 pos, float4 tex, float4 jac, float4 m0, float4 m1, float4 m3, float2 dim, thread float2& vpos)

#endif

{
	float2 n = normalize(pos.zw);
	float s = dot(m3.xy, pos.xy) + m3.w;
	float t = dot(m3.xy, n);

	float u = (s * dot(m0.xy, n) - t * (dot(m0.xy, pos.xy) + m0.w)) * dim.x;
	float v = (s * dot(m1.xy, n) - t * (dot(m1.xy, pos.xy) + m1.w)) * dim.y;

	float s2 = s * s;
	float st = s * t;
	float uv = u * u + v * v;
	float2 d = pos.zw * (s2 * (st + sqrt(uv)) / (uv - st * st));

	vpos = pos.xy + d;
	return (float2(tex.x + dot(d, jac.xy), tex.y + dot(d, jac.zw)));
}

#if defined(VULKAN) || defined(GL_core_profile) || defined(GL_ES)

	layout(location = 0) in float4 attrib0;
	layout(location = 1) in float4 attrib1;
	layout(location = 2) in float4 attrib2;
	layout(location = 3) in float4 attrib3;
	layout(location = 4) in float4 attrib4;

	#if defined(VULKAN)

		layout(push_constant) uniform ParamStruct
		{
			float4 slug_matrix[4];
			float2 slug_viewport;
		};

		layout(location = 0) out float4 vresult_color;
		layout(location = 1) out float2 vresult_texcoord;
		layout(location = 2) flat out float4 vresult_banding;
		layout(location = 3) flat out int4 vresult_glyph;

	#else

		uniform float4 slug_matrix[4];
		uniform float2 slug_viewport;

		out float4 vresult_color;
		out float2 vresult_texcoord;
		flat out float4 vresult_banding;
		flat out int4 vresult_glyph;

	#endif

	void main()
	{
		float2 p;

		vresult_texcoord = SlugDilate(attrib0, attrib1, attrib2, slug_matrix[0], slug_matrix[1], slug_matrix[3], slug_viewport, p);
		gl_Position.x = p.x * slug_matrix[0].x + p.y * slug_matrix[0].y + slug_matrix[0].w;
		gl_Position.y = p.x * slug_matrix[1].x + p.y * slug_matrix[1].y + slug_matrix[1].w;
		gl_Position.z = p.x * slug_matrix[2].x + p.y * slug_matrix[2].y + slug_matrix[2].w;
		gl_Position.w = p.x * slug_matrix[3].x + p.y * slug_matrix[3].y + slug_matrix[3].w;

		SlugUnpack(attrib1, attrib3, vresult_banding, vresult_glyph);
		vresult_color = attrib4;
	}

#elif defined(__METAL_VERSION__)

	struct AttribStruct
	{
		float4 position [[attribute(0)]];
		float4 texcoord [[attribute(1)]];
		float4 jacobian [[attribute(2)]];
		float4 banding [[attribute(3)]];
		float4 color [[attribute(4)]];
	};

	struct ParamStruct
	{
		float4 slug_matrix[4];
		float2 slug_viewport;
	};

	struct VertexStruct
	{
		float4 position [[position]];
		float4 color;
		float2 texcoord;
		float4 banding [[flat]];
		int4 glyph [[flat]];
	};

	vertex VertexStruct VertexShader(AttribStruct attrib [[stage_in]], constant ParamStruct *paramData [[buffer(1)]])
	{
		float2 p;
		VertexStruct vresult;

		vresult.texcoord = SlugDilate(attrib.position, attrib.texcoord, attrib.jacobian, paramData->slug_matrix[0], paramData->slug_matrix[1], paramData->slug_matrix[3], paramData->slug_viewport, p);
		vresult.position.x = p.x * paramData->slug_matrix[0].x + p.y * paramData->slug_matrix[0].y + paramData->slug_matrix[0].w;
		vresult.position.y = p.x * paramData->slug_matrix[1].x + p.y * paramData->slug_matrix[1].y + paramData->slug_matrix[1].w;
		vresult.position.z = p.x * paramData->slug_matrix[2].x + p.y * paramData->slug_matrix[2].y + paramData->slug_matrix[2].w;
		vresult.position.w = p.x * paramData->slug_matrix[3].x + p.y * paramData->slug_matrix[3].y + paramData->slug_matrix[3].w;

		SlugUnpack(attrib.texcoord, attrib.banding, vresult.banding, vresult.glyph);
		vresult.color = attrib.color;
		return (vresult);
	}

#elif defined(__PSSL__)

	struct AttribStruct
	{
		DataBuffer attrib[5];
	};

	struct ParamStruct
	{
		float4 slug_matrix[4];
		float2 slug_viewport;
	};

	struct DataStruct
	{
		AttribStruct *attribData;
		ParamStruct *paramData;
	};

	struct VertexStruct
	{
		float4 position : S_POSITION;
		float4 color : U_COLOR;
		float2 texcoord : U_TEXCOORD;
		nointerp float4 banding : U_BANDING;
		nointerp int4 glyph : U_GLYPH;
	};

	VertexStruct main(DataStruct data : S_SRT_DATA, uint id : S_VERTEX_ID)
	{
		float2 p;
		VertexStruct vresult;

		vresult.texcoord = SlugDilate(data.attribData->attrib[0][id], data.attribData->attrib[1][id], data.attribData->attrib[2][id], data.paramData->slug_matrix[0], data.paramData->slug_matrix[1], data.paramData->slug_matrix[3], data.paramData->slug_viewport, p);
		vresult.position.x = p.x * data.paramData->slug_matrix[0].x + p.y * data.paramData->slug_matrix[0].y + data.paramData->slug_matrix[0].w;
		vresult.position.y = p.x * data.paramData->slug_matrix[1].x + p.y * data.paramData->slug_matrix[1].y + data.paramData->slug_matrix[1].w;
		vresult.position.z = p.x * data.paramData->slug_matrix[2].x + p.y * data.paramData->slug_matrix[2].y + data.paramData->slug_matrix[2].w;
		vresult.position.w = p.x * data.paramData->slug_matrix[3].x + p.y * data.paramData->slug_matrix[3].y + data.paramData->slug_matrix[3].w;

		SlugUnpack(data.attribData->attrib[1][id], data.attribData->attrib[3][id], vresult.banding, vresult.glyph);
		vresult.color = data.attribData->attrib[4][id];
		return (vresult);
	}

#else	// HLSL

	cbuffer ParamStruct : register(b0)
	{
		float4 slug_matrix[4];
		float2 slug_viewport;
	};

	struct VertexStruct
	{
		float4 position : SV_Position;
		float4 color : U_COLOR;
		float2 texcoord : U_TEXCOORD;
		nointerpolation float4 banding : U_BANDING;
		nointerpolation int4 glyph : U_GLYPH;
	};

	VertexStruct main(float4 attrib[5] : ATTRIB, uint vid : SV_VertexID)
	{
		float2 p;
		VertexStruct vresult;

		vresult.texcoord = SlugDilate(attrib[0], attrib[1], attrib[2], slug_matrix[0], slug_matrix[1], slug_matrix[3], slug_viewport, p);
		vresult.position.x = p.x * slug_matrix[0].x + p.y * slug_matrix[0].y + slug_matrix[0].w;
		vresult.position.y = p.x * slug_matrix[1].x + p.y * slug_matrix[1].y + slug_matrix[1].w;
		vresult.position.z = p.x * slug_matrix[2].x + p.y * slug_matrix[2].y + slug_matrix[2].w;
		vresult.position.w = p.x * slug_matrix[3].x + p.y * slug_matrix[3].y + slug_matrix[3].w;

		SlugUnpack(attrib[1], attrib[3], vresult.banding, vresult.glyph);
		vresult.color = attrib[4];
		return (vresult);
	}

#endif
